/******************************************************************************
 * program:     rasimg library                                                *
 * function:    Library for saving/loading special ftg format.                *
 * modul:       ras_ftg.cc                                                    *
 * licency:     GPL or LGPL                                                   *
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "types.h"
//#include "pictures.h"
#include "raster.h"

#include "bitstrea.c"

//#ifndef _MSC_VER
//  #define fpos_t long
//#endif


#ifndef SupportFTG
  #define SupportFTG 10
#endif

int SavePictureBMP(const char *Name,const Image &Img); ///!!!
extern "C" {
int loadstruct(FILE *F,const char *description, ...);

int savestruct(FILE *F,const char *description, ...);
}


//////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
void PeelPlane(Raster2DAbstract & obr,Raster2DAbstract & PlaneX,int PlaneNum)
{
int y,MaxY;

 //PlaneX.Erase();
 if(PlaneX.GetPlanes()!=1 || PlaneX.Size1D!=obr.Size1D) return;

 if(PlaneNum<0 || PlaneNum>=obr.GetPlanes() || obr.Data2D==NULL) return;
 //PlaneX.Create(obr.SizeX(),obr.SizeY(),1 | NoFill);

 MaxY=(obr.Size2D<PlaneX.Size2D)?obr.Size2D:PlaneX.Size2D;
 for(y=0;y<MaxY;y++)		// main row loop
    {
    obr.GetRowRaster(y)->Peel1Bit(PlaneX.GetRowRaster(y)->Data1D,PlaneNum);
    }
}
/////////////////////////////////////////////////////////////
void JoinPlane(Raster2DAbstract & PlaneX, Raster2DAbstract & Obr, BYTE PlaneNum)
{
unsigned y,MaxY;

 if(((PlaneNum<0)||(PlaneNum>=Obr.GetPlanes()))||(PlaneX.Data2D==NULL)) return;
 if(Obr.Data2D==NULL)
    {/*	//if no raster available, create it and hope that PlaneNum is a maximal plane
    *Obr=CreateRaster2D(PlaneX.Size1D,PlaneX.Size2D,PlaneNum);
    if(Obr.Data2D==NULL)*/
	return;
    }

 MaxY=(Obr.Size2D<PlaneX.Size2D)?Obr.Size2D:PlaneX.Size2D;
 for(y=0;y<MaxY;y++)		// hlavni radkova smycka
    {
    Obr.GetRowRaster(y)->Join1Bit(PlaneX.GetRowRaster(y)->Data1D,PlaneNum);
    }
}
//////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


//--------------------------FTG---------------------------
#ifdef SupportFTG

//This is file type identifier. It must be present on the begining of each file.
typedef struct
	{
	char Identif[4];
        WORD version;
	} ImgHeader;

//This is object descriptor. Each file consists of set of this objects.
typedef struct 
        {
	BYTE    type;
        DWORD   size;
	} ImgObject;

#define SizeImgObject 5

inline long LoadFtgObject(FILE *f, ImgObject &SU)
{
#if defined(__PackedStructures__)
 return(fread(&SU,1,sizeof(SU),f));
#else
 return(loadstruct(f,"bd", &SU.type, &SU.size));
#endif
}

inline long SaveFtgObject(FILE *f, ImgObject &SU)
{
#if defined(__PackedStructures__)
 return(fwrite(&SU,1,sizeof(SU),f));
#else
 return(savestruct(f,"bd", SU.type, SU.size));
#endif
}

//Here follows definitions for each file object record
typedef struct
	{
	WORD    SizeX;
	WORD    SizeY;
	BYTE    planes;
	BYTE    interpretation;
	BYTE    compression;
	} RasterData;

typedef struct
	{
	WORD    items;
	BYTE    typ;            // 2-word Packed; 3-24 bits; 6-word for R,G,B
	WORD    startItem;
	BYTE    destPlanes;
	} PaletteData;


inline long LoadFtgData(FILE *f, RasterData &SU)
{
#if defined(__PackedStructures__)
 return(fread(&SU,1,sizeof(SU),f));
#else
 return(loadstruct(f,"wwbbb",
	 &SU.SizeX, &SU.SizeY,
	 &SU.planes, &SU.interpretation, &SU.compression));
#endif
}

inline long SaveFtgData(FILE *f, RasterData &SU)
{
#if defined(__PackedStructures__)
 return(fwrite(&SU,1,sizeof(SU),f));
#else
 return(savestruct(f,"wwbbb",
	 SU.SizeX, SU.SizeY,
	 SU.planes, SU.interpretation, SU.compression));
#endif
}


//This structures are used for data handling.
class TImageStruct;

class TObjectReader             //reads all objects stored in file
        {
public: FILE *ImgFile;
        fpos_t StartPosition;   //if ==0 then object contain invalid data
        fpos_t NextObjPos;
        fpos_t ActualPosition;
        fpos_t LastObjPosition;         //dodelat!!!!!
        ImgObject ActualObj;

        TImageStruct *AssignedImg;

        TObjectReader(void);
        void Init(fpos_t Position);
        void Rewind(void);
        int ReadObject(fpos_t Position);
	};

class TImageStruct              //main file descriptor for handling file
        {
public: FILE *ImgFile;
        WORD version;

        TObjectReader RootReader;

        int AttachFile(const char *FileName);
        void DeattachFile(void);

        TImageStruct();
        ~TImageStruct(void);
	};

/* Methods of class TObjectReader */
TObjectReader::TObjectReader(void) 
{
  AssignedImg=NULL;
  memset(&StartPosition,0,sizeof(StartPosition));
  return;
}


//This call initialize all object. You can read objects only after this call.
//      you may set Position=0 for actual file position
void TObjectReader::Init(fpos_t Position)
{
 StartPosition=0;
 if(AssignedImg==NULL) return;
 if(AssignedImg->ImgFile==NULL) return;
 if(Position!=0) StartPosition=Position;
	    else fgetpos(AssignedImg->ImgFile,&StartPosition);
 NextObjPos=ActualPosition=StartPosition;
}

void TObjectReader::Rewind(void)
{
 NextObjPos=StartPosition;
}


int TObjectReader::ReadObject(fpos_t Position)
{
 if(StartPosition==0) return(-11);      //error - Object not initialized
 if(NextObjPos==0) return(-12);         //error - Object not initialized

 if(Position!=0) ActualPosition=Position; //correct actual position if it is known
 if(ActualPosition!=NextObjPos)
        {
        if(fsetpos(AssignedImg->ImgFile,&NextObjPos)!=0) return(-12);
        ActualPosition=NextObjPos;
        }
 if(feof(AssignedImg->ImgFile)) return(-13);

 NextObjPos=0;

 if(LoadFtgObject(AssignedImg->ImgFile,ActualObj)!=SizeImgObject)
	return(-12);    //error - cannot read object descriptor;

 ActualPosition+=SizeImgObject;
 NextObjPos = ActualPosition+ActualObj.size;
 return(ActualObj.type);
}



/* Methods of class TImageStruct */
inline TImageStruct::TImageStruct(void)
{version=0;ImgFile=NULL;RootReader.AssignedImg=this;return;}

inline TImageStruct::~TImageStruct(void)
{DeattachFile();}


int TImageStruct::AttachFile(const char *FileName)
{
ImgHeader Hdr;

 if(ImgFile!=NULL) fclose(ImgFile);
 ImgFile=fopen(FileName,"rb");
 if (ImgFile==NULL) return(ErrOpenFile);         //File cannot be opened
 fread(&Hdr,sizeof(Hdr),1,ImgFile);

 if(memcmp(Hdr.Identif,"F&Tg",4))
        {
        DeattachFile();
        return(-2);                     //Another file
        }
 if(Hdr.version>=0x200)
        {
//      printf("version %X",Hdr.version);
        DeattachFile();
        return(-3);                     //Unsupported version
        }
 version=Hdr.version;

 RootReader.Init(sizeof(Hdr));

 return(0);                             //File opened OK
}

void TImageStruct::DeattachFile(void)
{
 version=0;
 if (ImgFile==NULL) return;
 fclose(ImgFile);
 ImgFile=NULL;
}

/*------------Procedures for F&Tg compression/decompression------------------*/
#define  SchMask  0x7
#define   Schless_2D    1
#define   Schless_3D    2
#define   Schless_3DVar 3
#define  HufMask   0x18
#define  Huff         8
#define  HuffVar   0x10
#define  RemapPal  0x20
#define  GrayCode  0x40
BYTE FtGrCompression=Schless_3DVar+HuffVar ; //actual type of compression

#define CBufSize 2048		//Size of bitstream file cache. Must be >=4!

#define ProbeCells   10

typedef BYTE bHuf[MAX_CODES];
typedef BYTE TSmaskFull[256];


SBYTE Repeaters[8]  = {0,0,3,7,15,31,63,127};
SBYTE ZeroValues[8] = {0,0,2,6,14,30,62,126};
SBYTE Aditors[8]    = {0,0,1,3, 7,15,31, 63};


INLINE int GetBinPix(unsigned x,BYTE *b)
{
 b+=x >> 3;
//GetBinPix:=b^ and ($80 >>(x and $7));{}
 return(*b & (1 <<(x & 7)));
}

//This procedure set one bit in the bit array
INLINE void SetBinPix(WORD x,BYTE *b)
{
 b+=x >> 3;
//{ b^:=b^ or ($80 >>(x and $7));{}
 *b=*b | (1 << (x & 7));
}


/*This procedure return a size of the number coded in log growth code*/
unsigned SizeNum(unsigned n)
{
 if(n<=2)   return(2);
 if(n<=4)   return(3);
 if(n<=8)   return(5);
 if(n<=16)  return(7);
 if(n<=32)  return(9);
 if(n<=64)  return(11);
 if(n<=128) return(13);
 if(n<=256) return(15);
 if(n<=512) return(17);
 return(19);
}


//Selecting neighbourhood planes for RGB
void Setn3n4forRGB(Raster2DAbstract & obr,BYTE n,BYTE & n2plane,BYTE & n3plane,BYTE & n4plane)
{
//  position	  ( 1, 2, 3, 4, 5, 6, 7, 8,   9,10,11,12,13,13,15,16,  17,18,19,20,21,22,23,24);
const BYTE Re2[24]={ 2, 3, 4, 5, 6, 7, 8,16,  10,11,12,13,14,15,16,24,  18,19,20,21,22,23,24,255};
const BYTE Re3[24]={ 9,10,11,12,13,14,15,15,  11,12,13,14,15,16,22,23,  19,20,21,22,23,24,255,255};
const BYTE Re4[24]={17,18,19,20,21,22,23,24,  17,18,19,20,21,22,23,22,  20,21,22,23,24,255,255,255};

if ((n>obr.GetPlanes())||(n<=0)) return;
n2plane=n+1;
n3plane=n+2;
n4plane=n+3;
if((FtGrCompression & RemapPal) != 0)
    {
    if(obr.GetPlanes()==24)		//True Color RGB mode}
	{
	n2plane=Re2[n];
	n3plane=Re3[n];
	n4plane=Re4[n];
	}
   }
if(--n2plane>=obr.GetPlanes()) n2plane=255;
if(--n3plane>=obr.GetPlanes()) n3plane=255;
if(--n4plane>=obr.GetPlanes()) n4plane=255;
}

//This procedure decompress packed description of the estimator
void ExplodeBitFlags(WORD & numera,TSmaskFull & SmaskF,TSmaskFull & SmaskEx)
{
WORD ri,i;

  memset(SmaskEx,0,sizeof(SmaskF));
  for(i=0;i<=2047;i++)
	{
	ri=i;
	if((numera & 2048) != 0)  ri= ((ri & 0xF000) >> 1) | (ri & 0x7FF);
	if((numera & 1024) != 0)  ri= ((ri & 0xF800) >> 1) | (ri & 0x3FF);
	if((numera &  512) != 0)  ri= ((ri & 0xFC00) >> 1) | (ri & 0x1FF);
	if((numera &  256) != 0)  ri= ((ri & 0xFE00) >> 1) | (ri & 0x0FF);
	if((numera &  128) != 0)  ri= ((ri & 0xFF00) >> 1) | (ri & 0x07F);
	if((numera &   64) != 0)  ri= ((ri & 0xFF80) >> 1) | (ri & 0x03F);
	if((numera &   32) != 0)  ri= ((ri & 0xFFC0) >> 1) | (ri & 0x01F);
	if((numera &   16) != 0)  ri= ((ri & 0xFFE0) >> 1) | (ri & 0x00F);
	if((numera &    8) != 0)  ri= ((ri & 0xFFF0) >> 1) | (ri & 0x007);
	if((numera &    4) != 0)  ri= ((ri & 0xFFF8) >> 1) | (ri & 0x003);
	if((numera &    2) != 0)  ri= ((ri & 0xFFFC) >> 1) | (ri & 0x001);
	if((numera &    1) != 0)  ri= ((ri & 0xFFFE) >> 1);

	if( GetBinPix(ri,(BYTE *)SmaskF)!=0)
		SetBinPix(i,(BYTE *)SmaskEx);
	}

}


//This procedure compute an optimal estimator for the selected bit plane
long SchCount3d2x3a(Raster2DAbstract & obr,Raster2DAbstract & PlaneX,BYTE n,WORD & numera,TSmaskFull & SmaskF)
{
int x,y;
BYTE mask;
int i;
unsigned long *a;		//array[0..4095] of longint;
long koutecku3f;
BYTE *p0,*p1,*p2,*p3,*p4,*p5,*p6;
BYTE *plane2a,*plane2b,*plane3,*plane4;
WORD lpl;

long qq[ProbeCells+1];
long lm,lg;
WORD c,num;
BYTE qplanes;

BYTE n2plane,n3plane,n4plane;

 if ((n>obr.GetPlanes()) || (obr.Data2D==NULL)) return(0);

 a=(unsigned long *)malloc(4096*sizeof(unsigned long));
 lpl=(obr.Size1D+7) / 8;

 plane2a=(BYTE *)malloc(lpl);
 plane2b=(BYTE *)malloc(lpl);
 plane3=(BYTE *)malloc(lpl);
 plane4=(BYTE *)malloc(lpl);

 memset(plane2a,0,lpl);
 memset(plane2b,0,lpl);
 memset(plane3,0,lpl);
 memset(plane4,0,lpl);

 Setn3n4forRGB(obr,n,n2plane,n3plane,n4plane);

 if(n2plane<=obr.GetPlanes()) obr.GetRowRaster(0)->Peel1Bit(plane2a,n2plane);

 memset(a,0,4096*sizeof(unsigned long));

 //FILE *f=fopen("c:\\temp\\aa.txt","wt"); //!!!!

 for(y=0;y<=obr.Size2D-2;y++)		//main row loop
   {
   if(n<obr.GetPlanes())
	{
	p3=plane2a;plane2a=plane2b;plane2b=p3;
	obr.GetRowRaster(y+1)->Peel1Bit(plane2a,n2plane);

	if(n+1<obr.GetPlanes())
	     {
	     obr.GetRowRaster(y+1)->Peel1Bit(plane3,n3plane);
	     if(n+2<obr.GetPlanes())
		{
		obr.GetRowRaster(y+1)->Peel1Bit(plane4,n4plane);
		}
	     }
	}

   p2=(BYTE *)PlaneX.GetRowRaster(y+1)->Data1D; // data[y+1].b;
   p1=(BYTE *)PlaneX.GetRowRaster(y)->Data1D;   //PlaneX.data[y].b;
   if(y==0) p0=p1;			//duplication of row 0
       else p0=(BYTE *)PlaneX.GetRowRaster(y-1)->Data1D; //y or y-1 ????

   p3=plane2b;
   p4=plane2a;

   p5=plane3;
   p6=plane4;

   i=0;
   if((*p0 & 128) != 0) i+=128;
   if((*p1 & 128) != 0) i+=512;
   if((*p2 & 128) != 0) i+=2048;
   if((*p3 & 128) != 0) i+=8;
   if((*p4 & 128) != 0) i+=32;

   mask=64;
   for(x=0;x<=obr.Size1D-2;x++)
	   {
	   i= ((i >> 1) & 0x554);
	   if ((*p0 & mask)!=0) i |= 128;
	   if ((*p1 & mask)!=0) i |= 512;
	   if ((*p2 & mask)!=0) i |=2048;
	   if ((*p3 & mask)!=0) i |=   8;
	   if ((*p4 & mask)!=0) i |=  32;
	   if ((*p5 & mask)!=0) i |=   2;
	   if ((*p6 & mask)!=0) i |=   1;

	   //fprintf(f,"%d:%d %d p0:%d\n",x,y,i,*p0);

	   mask=mask >> 1;
	   if(mask==0)
		   {
		   mask=128;
		   p0++;
		   p1++;
		   p2++;
		   p3++;
		   p4++;
		   p5++;
		   p6++;
		   }
	   a[i]++;
	   }
   }


free(plane2a);
free(plane2b);
free(plane3);
free(plane4);

/*unsigned long jj=0;
for(i=0;i<=4095;i++)
  {fprintf(f,"%d: %ld\n",i,a[i]);jj+=a[i];}
fprintf(f,"%ld",jj);
fclose(f);*/

//-----------vypocty diskriminativnosti smeru---------

qplanes=11;
if(numera==0)
   {
RepeatReduction:
   num=1 << qplanes;

   koutecku3f=0;
   for(i=0;i<=num-1;i++)
	  if(a[i+num]>a[i]) koutecku3f+=a[i];
		       else koutecku3f+=a[i+num];

   memset(qq,0,sizeof(qq));
   for(x=0;x<=qplanes-1;x++)
     {
     c=1 << x;

     for(i=0;i<=num-1;i++)
	  {
	  lm=a[i]+a[i ^ c];
	  lg=a[i + num]+a[(i ^ c) + num];

	  if(lm<lg) qq[x]+=lm;
	       else qq[x]+=lg;
	  }
     }

     x=0;		//find the direction with less diskriminativness
     lm=0;
     for(i=qplanes-1;i>=0;i--)
	  {
	  qq[i]=(qq[i] >> 1) - koutecku3f;
	  if(lm>=qq[i])
		  {
		  lm=qq[i];
		  x=i;
		  }
	  }

		// bits of the table

     if((num > 2*qq[x])&&(qplanes>=1))
	  {				//remove direction i
	  c=1 << x;

	  y=0;
	  for(i=0;i<=(2 * num)-1;i++)
		  {
		  if((i & c) != 0) continue;
		  a[y]=a[i]+a[i | c];
		  y++;
		  }

	  for(i=0;i<=15;i++)
	       {
	       if(( (numera&(1<<i))!=0 )&&(x>=i)) x++;
	       }
	  numera |= 1 << x;
	  qplanes--;

	  goto RepeatReduction;
	  }
     }
     else {
	  for(x=qplanes-1;x>=0;x--)
		{
		num=1 << qplanes;
		if( (numera&(1<<(x-1)))!=0 )  //x in numera
			{
			c=1 << x;

			y=0;
			for(i=0;i<=(2 * num)-1;i++)
				{
				if((i & c) != 0) continue;
				a[y]=a[i]+a[i | c];
				y++;
				}
			qplanes--;
			}
		}
	  num=1 << qplanes;
	  }


memset(SmaskF,0,sizeof(SmaskF));
koutecku3f=0;
for(i=0;i<num;i++)
	{
	if(a[i+num]>a[i]) {
			  SetBinPix(i,(BYTE *)SmaskF);
			  koutecku3f+=a[i];
			  }
		     else koutecku3f+=a[i+num];
	}

free(a);
return(koutecku3f);
}


// perform compression of the selected bit plane and known predictor
void SchComp3d2x3a(Raster2DAbstract & obr,Raster2DAbstract & PlaneC,BYTE n,WORD numera,TSmaskFull & SmaskF)
{
int i,x,y;
BYTE Rmask,OldRmask;
BYTE *p0,*p1,*p2,*p3,*p4,*p5,*p6;
BYTE *plane2a,*plane2b,*plane3a,*plane3b,*plane4a,*plane4b;
WORD lpl;
TSmaskFull SmaskEx;
BYTE n2plane,n3plane,n4plane;

 if(n>obr.GetPlanes() || obr.Data2D==NULL) return;
 if(PlaneC.GetPlanes()!=1 || PlaneC.Data2D==NULL) PeelPlane(obr,PlaneC,n-1);

 ExplodeBitFlags(numera,SmaskF,SmaskEx);
 lpl=(obr.Size1D+7) / 8;

 plane2a=(BYTE *)malloc(lpl);
 plane2b=(BYTE *)malloc(lpl);
 plane3a=(BYTE *)malloc(lpl);
 plane3b=(BYTE *)malloc(lpl);
 plane4a=(BYTE *)malloc(lpl);
 plane4b=(BYTE *)malloc(lpl);

 memset(plane2a,0,lpl);
 memset(plane2b,0,lpl);
 memset(plane3a,0,lpl);
 memset(plane3b,0,lpl);
 memset(plane4a,0,lpl);
 memset(plane4b,0,lpl);

 Setn3n4forRGB(obr,n,n2plane,n3plane,n4plane);

 obr.GetRowRaster(obr.Size2D-1)->Peel1Bit(plane2b,n2plane);
 obr.GetRowRaster(obr.Size2D-1)->Peel1Bit(plane3b,n3plane);
 obr.GetRowRaster(obr.Size2D-1)->Peel1Bit(plane4b,n4plane);

 for(y=obr.Size2D-2;y>=0;y--)		//main rows loop
	{
	if(n<obr.GetPlanes())
		{
		p3=plane2a;plane2a=plane2b;plane2b=p3;
		obr.GetRowRaster(y)->Peel1Bit(plane2b,n2plane);

		if(n+1<obr.GetPlanes())
		     {
		     p5=plane3a;plane3a=plane3b;plane3b=p5;
		     obr.GetRowRaster(y)->Peel1Bit(plane3b,n3plane);
		     if(n+2<obr.GetPlanes())
			{
			p6=plane4a;plane4a=plane4b;plane4b=p6;
			obr.GetRowRaster(y)->Peel1Bit(plane4b,n4plane);
			}
		     }
		}

	p2=(BYTE *)PlaneC.GetRowRaster(y+1)->Data1D;
	p1=(BYTE *)PlaneC.GetRowRaster(y)->Data1D;
	if(y==0) p0=p1;
	    else p0=(BYTE *)PlaneC.GetRowRaster(y-1)->Data1D;


	p3=plane2b;
	p4=plane2a;

	p5=plane3a;
	p6=plane4a;

	p0+=(obr.Size1D-1) >> 3;
	p1+=(obr.Size1D-1) >> 3;
	p2+=(obr.Size1D-1) >> 3;
	p3+=(obr.Size1D-1) >> 3;
	p4+=(obr.Size1D-1) >> 3;
	p5+=(obr.Size1D-1) >> 3;
	p6+=(obr.Size1D-1) >> 3;

	Rmask=128 >> ((obr.Size1D-1) & 7);

	i=0;
	if( (*p0 & Rmask) != 0) i+=64;
	if( (*p1 & Rmask) != 0) i+=256;	//obr.Size1D-1 column
	if( (*p2 & Rmask) != 0) i+=1024;
	if( (*p3 & Rmask) != 0) i+=4;
	if( (*p4 & Rmask) != 0) i+=16;

	OldRmask=Rmask;
	Rmask=Rmask << 1;
	if(Rmask==0) {
		     Rmask=1;
		     p0--;
		     p1--;
		     p2--;
		     p3--;
		     p4--;
		     }

	for(x=obr.Size1D-2;x>=0;x--)
		{
		i= ((i << 1) & 0xAA8);
		if((*p0 & Rmask)!=0)  i |=  64;
		if((*p1 & Rmask)!=0)  i |= 256;
		if((*p2 & Rmask)!=0)  i |=1024;
		if((*p3 & Rmask)!=0)  i |=   4;
		if((*p4 & Rmask)!=0)  i |=  16;

		if((*p5 & OldRmask)!=0)	i |= 2;
		if((*p6 & OldRmask)!=0) i |= 1;

		OldRmask=Rmask;
		if(OldRmask==1)
		       {
		       p5--;
		       p6--;
		       }
		Rmask=Rmask << 1;
		if(Rmask==0)
			{
			Rmask=1;
			p0--;
			p1--;
			p2--;
			p3--;
			p4--;
			}


		if ((GetBinPix(i & 2047,(BYTE *)SmaskEx)!=0))
			{
			if (i>=2048)  PlaneC.SetValue2D(x+1,y+1,0); //px=1
				 else PlaneC.SetValue2D(x+1,y+1,1); //px=0
			}

		}
	}

free(plane2a);
free(plane2b);
free(plane3a);
free(plane3b);
free(plane4a);
free(plane4b);
}



//RLC Compression of the 1st row and 1st column
unsigned Schless1d(Raster2DAbstract & obr,int x,int y,unsigned kx,unsigned ky,BYTE & S1)
{
BYTE M;
int i,Max;
unsigned a[4];
unsigned koutecku;

 if(obr.Data2D==NULL || obr.GetPlanes()!=1) return(0);

 memset(a,0,sizeof(a)); // pozor !!!

 if(kx==0) Max=obr.Size2D-2;
 if(ky==0) Max=obr.Size1D-2;

 for(i=0;i<=Max;i++)
	{
	M=   obr.GetValue2D(x+(kx & i),y+(ky & i))+
	   2*obr.GetValue2D(x+(kx & (i+1)),y+(ky & (i+1)));
	a[M]++;
	}

 S1=0;
 koutecku=0;

 for(i=0;i<=1;i++)
	{
	if(a[i+2]>a[i])
		{
		S1=S1 | (1 << i);
		koutecku+=a[i];
		}
		else koutecku+=a[i+2];
	}


 for(i=Max;i>=0;i--)
	{
	M=obr.GetValue2D(x+(kx & i),y+(ky & i));

	if( ( (S1 & (1 << M))!=0 ) ^ (obr.GetValue2D(x+(kx & (i+1)),y+(ky & (i+1)))==1))
		     obr.SetValue2D(x+(kx & (i+1)),y+(ky & (i+1)),1);
		else obr.SetValue2D(x+(kx & (i+1)),y+(ky & (i+1)),0);
	}

return(koutecku);
}


//RLC Decompression of the 1st row and 1st column
void DeSchless1d(Raster2DAbstract & obr,WORD x,WORD y,WORD kx,WORD ky,BYTE S1)
{
BYTE M;
WORD Max;
WORD i;

 if(kx==0) Max=obr.Size2D-2;
 if(ky==0) Max=obr.Size1D-2;

 for(i=0;i<=Max;i++)
	{
	M=obr.GetValue2D(x+(kx & i),y+(ky & i));

	if ( (((1<<M) & S1)!=0) ^ (obr.GetValue2D(x+(kx & (i+1)),y+(ky & (i+1)))==1))
		     obr.SetValue2D(x+(kx & (i+1)),y+(ky & (i+1)),1);
		else obr.SetValue2D(x+(kx & (i+1)),y+(ky & (i+1)),0);
	}
}


//This procedure performs a decompression of the selected compressed bit plane and known predictor
void DeSchless3d2x3a(Raster2DAbstract & obr,Raster2DAbstract & PlaneC,BYTE n,WORD & numera,TSmaskFull & SmaskF)
{
WORD x,y;
BYTE mask;
WORD i;
BYTE *p0,*p1,*p2,*p3,*p4,*p5,*p6;
void *plane2a,*plane2b,*plane3,*plane4;
WORD lpl;
TSmaskFull SmaskEx;
BYTE n2plane,n3plane,n4plane;

 if(n>obr.GetPlanes() || obr.Data2D==NULL || PlaneC.Data2D==NULL) return;
 ExplodeBitFlags(numera,SmaskF,SmaskEx);

 lpl=(obr.Size1D+7) / 8;
 plane2a=malloc(lpl);
 plane2b=malloc(lpl);
 plane3= malloc(lpl);
 plane4= malloc(lpl);

 memset(plane2a,0,lpl);
 memset(plane2b,0,lpl);
 memset(plane3,0,lpl);
 memset(plane4,0,lpl);

 Setn3n4forRGB(obr,n,n2plane,n3plane,n4plane);

 obr.GetRowRaster(0)->Peel1Bit(plane2a,n2plane);

 for(y=0;y<=obr.Size2D-2;y++)	//main loop through rows
	{
	if(n<obr.GetPlanes())
		{
		if((numera & (4|8|16|32)) != (4|8|16|32))
		  {
		  p3=(BYTE *)plane2a;plane2a=plane2b;plane2b=p3;
		  obr.GetRowRaster(y+1)->Peel1Bit(plane2a,n2plane);
		  }

		if(n+1<obr.GetPlanes())
		     {
		     if((numera & 2)!=2)
			{
			obr.GetRowRaster(y+1)->Peel1Bit(plane3,n3plane);
			}

		     if(n+2<obr.GetPlanes())
			{
			if((numera & 1)!=1)
			    {
			    obr.GetRowRaster(y+1)->Peel1Bit(plane4,n4plane);
			    }
			}
		     }
		}


	p2=(BYTE *)PlaneC.GetRowRaster(y+1)->Data1D;
	p1=(BYTE *)PlaneC.GetRowRaster(y)->Data1D;
	if(y==0) p0=p1;
	    else p0=(BYTE *)PlaneC.GetRowRaster(y-1)->Data1D;

	p3=(BYTE *)plane2b;
	p4=(BYTE *)plane2a;

	p5=(BYTE *)plane3;
	p6=(BYTE *)plane4;


	i=0;
	if( (*p0 & 128) != 0 ) i+=128;
	if( (*p1 & 128) != 0 ) i+=512;
	if( (*p2 & 128) != 0 ) i+=2048;
	if( (*p3 & 128) != 0 ) i+=8;
	if( (*p4 & 128) != 0 ) i+=32;

	mask=64;

	for(x=0;x<=obr.Size1D-2;x++)
		{
		i= ((i >> 1) & 0x554);
		if( (*p0 & mask)!=0 ) i=i |  128;
		if( (*p1 & mask)!=0 ) i=i |  512;
		if( (*p2 & mask)!=0 ) i=i | 2048;
		if( (*p3 & mask)!=0 ) i=i |    8;
		if( (*p4 & mask)!=0 ) i=i |   32;
		if( (*p5 & mask)!=0 ) i=i |    2;
		if( (*p6 & mask)!=0 ) i=i |    1;

		if ((GetBinPix(i  &  2047,SmaskEx)!=0))
			{
			i=i ^ 2048;
			*p2=*p2 ^ mask;
			}

		mask=mask >> 1;
		if(mask==0)
			{
			mask=128;
			p0++;
			p1++;
			p2++;
			p3++;
			p4++;
			p5++;
			p6++;
			}

		}
	}


free(plane2a);
free(plane2b);
free(plane3);
free(plane4);
}


//this procedure is used for compression of the huffman description table
//prefixes - used prefixes inside number stream
unsigned PackSizes(bits_t *Bit,BYTE *b,WORD MaxItems)
{
int i,j;
short x,Max,MaxDif,t,lastS;
unsigned FinalSize;

FinalSize=0;

lastS=2;
Max=b[0];
MaxDif=0;

for(i=1;i<=MaxItems;i++)
	{
	t=b[i];
	if(t!=0) {
		 if((abs(lastS-t))>MaxDif) MaxDif=abs(lastS-t);
		 lastS=t;
		 }
	if(t>Max) Max=t;
        }

t=127;		//zatim neumim
if(Max<=30) t= 2;  // primy kod 0..30 RLC 31
if(Max<=14) t= 1;  // primy kod 0..14 RLC 15
if(Max<= 6) t= 0;  // primy kod 0.. 6 RLC  7

if((MaxDif<= 6)&&(Max>6)) t=5;
if (MaxDif<= 2)		  t=4;

if(t==127) return(0);		//I cannot store


if(Bit!=NULL) sq_wrn(Bit,t,3);
FinalSize+=3;


i=0;
Max=(t & 3) + 3;
if(t>=4)
     {			// Store histogram diff accepted
     lastS=2;

     if(Bit!=NULL)  WriteNumber(Bit,b[0]+1);
     FinalSize+=SizeNum(b[0]+1);

     for(i=1;i<=MaxItems;i++)
	{
	x=b[i];
	if(x==0) b[i]=ZeroValues[Max];
	    else {
		 b[i]=x-lastS+Aditors[Max];
		 lastS=x;
		 }
	}

      i=1;
      }


lastS=2;
while(i<=MaxItems)	// Store histogram RLC; no diff in this place
	{
	x=b[i];

	j=i;
	if(x==lastS)
		{
		while ((j<=MaxItems)&&(b[j]==lastS))
			{
			j++;
			}
		}


	if(j-i>1 ) 		//komprese se vyplati pro opakovani >= 3x
		 {
		 if(Bit!=NULL)
			{
			sq_wrn(Bit,Repeaters[Max],Max);
			WriteNumber(Bit,j-i-1);
			}
		 FinalSize+=Max;
		 FinalSize+=SizeNum(j-i-1);
		 i=j-1;
		 }
		else
		{
		if(Bit!=NULL)  sq_wrn(Bit,x,Max);
		FinalSize+=Max;
		}

	lastS=x;
	i++;
	}


if(t>=4)		// inverse diff process
     {
     lastS=2;
     for(i=1;i<=MaxItems;i++)
	      {
	      if(b[i]==ZeroValues[Max]) b[i]=0;
	      else
		{
		b[i]-=Aditors[Max];
		b[i]+=lastS;
		lastS=b[i];
		}
	      }
     }

return(FinalSize);
}


//This procedure is used for decompression of the huffman description table
void UnPackSizes(bits_t *Bit,BYTE *b,unsigned MaxItems, short t)
{
int i,j;
short Vtyp,typ,lastS;

typ=sq_rdn(Bit,3);
i=0;
j=0;

if(typ>=4)
	{
	b[0]=ReadNumber(Bit)-1;
	i=1;
	}
lastS=2;

Vtyp=(typ & 3) + 3;
while(i<=MaxItems)
	 {
//	 b[i].Number=i;

	 if(j>0)
	       {
	       j--;
	       b[i]=lastS;
	       i++;
	       continue;
	       }

	 b[i]=sq_rdn(Bit,Vtyp);
	 if(b[i]==Repeaters[Vtyp])
		   {
		   b[i]=lastS;
		   j=ReadNumber(Bit);
		   i++;
		   continue;
		   }
	 lastS=b[i];
	 i++;
	 }


if(typ>=4)
    {
    lastS=2;
    for(i=1;i<=MaxItems;i++)
	      {
	      if(b[i]==ZeroValues[Vtyp]) b[i]=0;
		       else
		       {
		       b[i]-=Aditors[Vtyp];
		       b[i]=lastS+b[i];
		       lastS=b[i];
		       }
	      }
    }
}


/*This procedure evaluates huffman codes*/
int HufCount(huf_wr_t *b, unsigned HItems)
{
unsigned i;
long x,y;
unsigned BitHistogram[17];

 memset(BitHistogram,0,sizeof(BitHistogram));

 for(i=0;i<HItems;i++)
   BitHistogram[b->ln[i]]++;

 y=0;
 BitHistogram[0]=0;
 for(i=1;i<=16;i++)
   {
   x = BitHistogram[i];

   if(x!=0) BitHistogram[i] = y;
   y <<= 1;
   y+=x;
   }

/* Fill calculated Huffman codes
 for(i=0; i<HItems; i++)
   {
   if(b->ln[i]==0) b->cod[i] = 0;
	      else b->cod[i] = swap_bits_order(BitHistogram[b->ln[i]],b->ln[i]);
   BitHistogram[b->ln[i]]++;
   }
*/

 return y==65536l;
}


#if SupportFTG>=4 || SupportFTG==2
Image LoadPictureFTG(const char *Name)
{
TImageStruct NewImage;
TObjectReader *ActualReader;
RasterData RDataDesc;
PaletteData PaletteDesc;
int i,x,y;
int ldblk;
bits_t BitStream;
huf_rd_t *phuf;
BYTE *CompBuffer=NULL;
WORD renums;
TSmaskFull mFull;
BYTE Smask1dx,Smask1dy;
BYTE isHuff,HufN,HufTblSize;
BYTE *bha=NULL;
long NewX;
Image p;
Raster2DAbstract *RasRead=NULL;
Raster2DAbstract *PlaneX;
APalette *Palette=NULL;

 if((i=NewImage.AttachFile(Name))!=0) return(p);
 ActualReader=&NewImage.RootReader;

//printf("File opened\n\r");

 while(ActualReader->ReadObject(ActualReader->NextObjPos)>=0)
   {
//	printf("Object type %d, size %d\n\r",ActualReader->ActualObj.type,ActualReader->ActualObj.size);
   switch(ActualReader->ActualObj.type)
   {
     case 4:{
            if(LoadFtgData(NewImage.ImgFile,RDataDesc)!=7) break;

//          printf("  Raster Image:: Size1D:%d  sizeY:%d  planes:%d  compression:%d \n\r",RDataDesc.Size1D,RDataDesc.SizeY,RDataDesc.GetPlanes(),RDataDesc.compression);

	    RasRead = CreateRaster2D(RDataDesc.SizeX,RDataDesc.SizeY,RDataDesc.planes);
            if(RasRead==NULL) break;
	    if(RasRead->Data2D==NULL) break;

//	    printf("         OK to load & Init memory struct\n\r");

	    if(RDataDesc.compression==0)
	      {
	      ldblk=(int)(((long)RDataDesc.SizeX*RDataDesc.planes+7)/8);
	      for(y=0;y<RDataDesc.SizeY;y++)
		{
		if(fread(RasRead->GetRowRaster(y),1,ldblk,NewImage.ImgFile)!=ldblk)
		  {
		  p.Erase();
		  break;
		  }
		}
	   } else
	     {
	     ldblk = (RasRead->GetSize1D()+7) / 8;

	     bha=(BYTE *)malloc(sizeof(bHuf));
	     phuf=(huf_rd_t *)malloc(sizeof(huf_rd_t));
	     CompBuffer=(BYTE *)malloc(CBufSize);

	     if(CompBuffer==NULL || bha==NULL || phuf==NULL)
	       {
	       p.Erase();
	       return(p);					}

								/* - sizeof struct */
	       sq_rdinit(&BitStream,CompBuffer,CBufSize,ActualReader->ActualObj.size,NewImage.ImgFile);

	       for(i=RasRead->GetPlanes();i>=1;i--)
		 {
		 PlaneX = CreateRaster2D(RasRead->GetSize1D(),RasRead->Size2D,1);
		 PlaneX->Cleanup();	//fill zeros to all data

		 if(sq_rdn(&BitStream,1))
		   {			//plane is compressed
		   if((RDataDesc.compression & SchMask) < Schless_3DVar)
		     {
		     if(sq_rdn(&BitStream,1)) renums=1 | 2 | 64 | 128;
		                         else renums=1 | 2 | 4 | 8 | 16 | 32 | 64 | 128;
		     }
		   else renums = sq_rdn(&BitStream,16);

		   y=0;
		   for(x=1; x<=(1<<ProbeCells); x<<=1) 
		     if(!(x & renums)) y++;

		   memset(mFull,0,sizeof(mFull));
		   sq_rdUnl(&BitStream,mFull,1 << y);

		   Smask1dx=sq_rdn(&BitStream,2);
		   Smask1dy=sq_rdn(&BitStream,2);

		   if((isHuff=sq_rdn(&BitStream,1))==1)
		     {
		     HufN = sq_rdn(&BitStream,3);
		     HufTblSize = 0;

		     UnPackSizes(&BitStream,bha,Hu2Sizes[HufN],HufTblSize);

		     x=Hu2Sizes[HufN]+1;
		     bha[x]=0xFF;

		   if(!sq_rdhufi(phuf,x,8,bha))
		     {
		     return(p);
		     }

		   HufN+=Hu2Prefix[0]; 	//prefix length in bites
		   }


		 NewX=-1;
		 y=0;
		 while(y<PlaneX->Size2D)
		   {
		   if(isHuff==1) NewX+=ReadHufNum(&BitStream,phuf,HufN);
			    else NewX+=ReadNumber(&BitStream);

		   if(NewX>=PlaneX->GetSize1D()-1)
		     {
		     while(NewX>PlaneX->GetSize1D()-1)		//end of row
		       {
		       y++;
		       NewX-=PlaneX->GetSize1D();
		       }

		   if(NewX==PlaneX->GetSize1D()-1 && y==PlaneX->Size2D-1)
		     {
		     PlaneX->SetValue2D((int)NewX,y,1);   //last pixel
		     break;
		     }
		   }
		 if(y<=PlaneX->Size2D-1) PlaneX->SetValue2D((int)NewX,y,1);
		 }

		 if((NewX!=0)&&(NewX!=PlaneX->GetSize1D()-1))
		   {
		   printf("error NewX:%d!=%d\n",NewX,PlaneX->GetSize1D()-1);
		   goto ExitLoop;
		   //free(CompBuffer);
		   //free(bha);
		   //free(phuf);
		   //return(p);
		   }

//!!!!!!!!
//		char *Name=strdup("c:\\temp\\0_.bmp"); Name[8]='0'+i;
//		PlaneX->UsageCount=1;
//		SavePictureBMP(Name,PlaneX); 
//		free(Name);	///!!!!!!!!!!!!!!!!!!!!!!

		 DeSchless1d(*PlaneX,0,0,0xFFFF,0,Smask1dx);
		 DeSchless1d(*PlaneX,0,0,0,0xFFFF,Smask1dy);

		 DeSchless3d2x3a(*RasRead,*PlaneX,i,renums,mFull);
		 }
	       else for(y=0;y<RasRead->Size2D;y++)	//plane is stored uncompressed
	         {
	         sq_rdUnl(&BitStream,(BYTE *)PlaneX->GetRowRaster(y)->Data1D,RasRead->GetSize1D());
	         x=(PlaneX->GetSize1D()-1)>>3;
	         ((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]=swap_bits_xlat[((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]];
	         }

	       JoinPlane(*PlaneX,*RasRead,i-1);
	       }

	     free(CompBuffer);CompBuffer=NULL;
	     free(bha); bha=NULL;
	     free(phuf);phuf=NULL;
	     delete PlaneX;PlaneX=NULL;
	     }


	   break;
	   }
     case 20:
           {
	   if(fread(&PaletteDesc,sizeof(PaletteData),1,NewImage.ImgFile)!=1) break;

//	   printf("  Palette:: items:%d  typ:%d  planes:%d\n\r",PaletteDesc.items,PaletteDesc.typ,PaletteDesc.destPlanes);

	   if(PaletteDesc.items!=0)
	     {
	     Palette = BuildPalette(PaletteDesc.items, PaletteDesc.destPlanes);
	     if(!Palette) break;

	     for(i=0;i<Palette->Size1D;i++)
		{
		Palette->R(i,fgetc(NewImage.ImgFile));
		Palette->G(i,fgetc(NewImage.ImgFile));
		Palette->B(i,fgetc(NewImage.ImgFile));
		}
	     }
	   break;
	   }
     }
   }

ExitLoop:
 if(bha!=NULL) free(bha);
 if(PlaneX) {delete PlaneX;PlaneX=NULL;}
 if(RasRead!=NULL)
    {p.Raster=RasRead; p.Raster->UsageCount++;}
 if(Palette!=NULL)
    {p.Palette=Palette;p.Palette->UsageCount++;}
return(p);
}
#endif


#if SupportFTG>=3
int SavePictureFTG(const char *Name,const Image & p)
{
ImgHeader Header;
ImgObject ObjLeader;
RasterData RDataDesc;
PaletteData PaletteDesc;
unsigned ldblk;
bits_t BitStream;
BYTE *CompBuffer=NULL;
WORD renums;
long koutecku,lastX;
BYTE Smask1d;
TSmaskFull mFull;
BYTE isHuffman,HufN;
BYTE Rotm,*pb;
count_t *Histogram=NULL;
count_t HufZero,HufMax;
huf_wr_t *HufStruct=NULL;
ch_tab_t *ch_tab=NULL;
int MaxHuTbSize;
unsigned BcountLen[MAX_BITS];
unsigned StartPos;

FILE *f;
int SavePictureFTG_;
int i,x,y;
Raster2DAbstract *RasWr;
Raster2DAbstract *PlaneX=NULL;

 RasWr=p.Raster;
 if(RasWr==NULL) return(-10);
 if(RasWr->Data2D==NULL) return(-11);

 SavePictureFTG_=0;
 if((f=fopen(Name,"wb"))==NULL) return(ErrOpenFile);

 memmove(Header.Identif,"F&Tg",4);	//write image Header to disk
 Header.version=2;
 fwrite(&Header,sizeof(Header),1,f);

 StartPos = ftell(f);
 ObjLeader.type = 4;		//=image;
 ObjLeader.size = ~1;
 if(SaveFtgObject(f,ObjLeader)!=SizeImgObject) return(-12); //error - cannot write object descriptor

 memset(&RDataDesc,0,sizeof(RDataDesc));
 RDataDesc.SizeX=RasWr->GetSize1D();	//Rows
 RDataDesc.SizeY=RasWr->Size2D;		//Cols
 RDataDesc.planes=RasWr->GetPlanes();
 RDataDesc.compression=FtGrCompression;
 if((RasWr->GetSize1D()<8)&&(RasWr->Size2D<8)) RDataDesc.compression=0;

 SaveFtgData(f,RDataDesc);

/*  !!!!!!!!!!!!!!! DODELAT !!!!!!!!!!!!!!!!
 if(ImgRec.Compression!=0)
   {
   if(ImgRec.Compression & ReMapPal)!=0)
      {
     if(p.palette!=NULL) ReMapPalette(p)
     else
	else if (ImgRec.Compression & GrayCode)!=0 then Bin2Gray(p);
	 end;
   } */

 ldblk=(RasWr->GetPlanes()*RasWr->GetSize1D()+7) / 8;

 if(RDataDesc.compression==0)
    {
    for(y=0;y<RasWr->Size2D;y++)			//no compression mode
	{
	fwrite(RasWr->GetRowRaster(y)->Data1D,ldblk,1,f);
//	AlineProc(y,p);
	}
    }
    else {
	 CompBuffer=(BYTE *)malloc(CBufSize);
	 sq_wrinit(&BitStream,CompBuffer,CBufSize,0xFFFFFFFl,f);

	 PlaneX=CreateRaster2D(RasWr->GetSize1D(),RasWr->Size2D,1);
	 if(PlaneX==NULL) return(-13);

	 for(i=RasWr->GetPlanes();i>=1;i--)
	    {
	    PeelPlane(*RasWr,*PlaneX,i-1);
	    if(PlaneX->Data2D==NULL) return(-13);

	    renums=0;
	    if ((RDataDesc.compression & SchMask) < Schless_3D) renums=1 | 2 | 4 | 8 | 16 | 32 | 64 | 128;
			else {
			     if((RDataDesc.compression & SchMask) < Schless_3DVar)
				renums=1 | 2 | 64 | 128;
			     }

	    koutecku=SchCount3d2x3a(*RasWr,*PlaneX,i,renums,mFull)+1;

	    y=0;
	    for(x=1;x<=(1 << ProbeCells);x<<=1) if(!(x & renums)) y++;

	    if( ((2.5*koutecku)+(1 << y)) < ((long)(PlaneX->GetSize1D()-1))*(PlaneX->Size2D-1))
		{		//is it a good think to compress this plane?
		sq_wrn(&BitStream,1,1);

		if ((RDataDesc.compression & SchMask) < Schless_3DVar)
			{
			if(renums==(1 | 2 | 4 | 8 | 16 | 32 | 64 | 128))
					sq_wrn(&BitStream,0,1);
				   else sq_wrn(&BitStream,1,1);
			}
			else sq_wrn(&BitStream,renums,16);

		 sq_wrUnl(&BitStream,mFull,1 << y);

		 SchComp3d2x3a(*RasWr,*PlaneX,i,renums,mFull);

		 Schless1d(*PlaneX,0,0,0xFFFF,0,Smask1d); 	//0-th row
		 sq_wrn(&BitStream,Smask1d,2);
		 Schless1d(*PlaneX,0,0,0,0xFFFF,Smask1d);	//0-th column
		 sq_wrn(&BitStream,Smask1d,2);

		 isHuffman=0;
//
//		char *Name="c:\\temp\\0_.bmp"; Name[8]='0'+i;
//		PlaneX->UsageCount=1;
//		SavePictureBMP(Name,PlaneX); ///!!!
//	 -------Huffmann compression----}
		 if((RDataDesc.compression & HufMask)!=0)
			{
			if(Histogram==NULL) Histogram=(count_t *)malloc(MAX_CODES*sizeof(count_t));
			if(HufStruct==NULL) HufStruct=(huf_wr_t *)malloc(sizeof(huf_wr_t));
			if(ch_tab==NULL)    ch_tab=(ch_tab_t *)malloc(MAX_CODES*sizeof(ch_tab));
			if((Histogram==NULL)||(HufStruct==NULL)||(ch_tab==NULL)) goto NoHuff;

			memset(Histogram,0,MAX_CODES*sizeof(count_t));
			lastX=-1;
			for(y=0;y<PlaneX->Size2D;y++)
			    {
			    Rotm=128;
			    pb=(BYTE *)PlaneX->GetRowRaster(y)->Data1D;

			    for(x=0;x<PlaneX->GetSize1D();x++)
				{
				if( (*pb & Rotm) != 0)
					{
					lastX=(long)(x)-lastX;
					if(lastX>=MAX_CODES) lastX=0;
					Histogram[(int)lastX]++;
					lastX=x;
					}

				Rotm=Rotm >> 1;
				if(Rotm==0)
					{
					Rotm=128;
					pb++;
					}
				}
			    lastX=lastX - PlaneX->GetSize1D();
			    }
			 if(lastX<0)
				{
				lastX= -lastX;
				if(lastX>=MAX_CODES) lastX=0;
				Histogram[(int)lastX]++;
				}

			HufZero=Histogram[0];
			HufMax=0;
			y=0;

			if( (RDataDesc.compression & HufMask)>=HuffVar )
			  {
			    /*Iterate through all supported splitting a number se to 2 groups */
			    /* first group is encoded by huffman and a second group is stored in logaritmical code */
			  for(HufN=0;HufN<=6;HufN++)
			    {
			    Histogram[0]=HufZero;
			    for(x=Hu2Sizes[HufN]+1;x<MAX_CODES;x++) Histogram[0]+=Histogram[x];

			    x=sq_huffman(Histogram,HufStruct->ln,BcountLen,Hu2Sizes[HufN]+1,ch_tab);

				/* Calculate data stream difference-size in bits. */
			    lastX = Histogram[0] * (Hu2Prefix[HufN] - HufStruct->ln[0]);

			    // 1st set is Huffman encoded and its elements ere Huffman encoded */
			    for(x=1;x<=Hu2Sizes[HufN];x++)
				  {
				  lastX += Histogram[x] * ((long)SizeNum(x) - HufStruct->ln[x]);
				  }			  //data length change

			    MaxHuTbSize=PackSizes(NULL,HufStruct->ln,Hu2Sizes[HufN]);
			    lastX -= MaxHuTbSize;  //- length of the table decreases criteria

				// when a new bitstream is longer by more than 18 bits, don't try next splitting
			    if( (lastX<HufMax+18) && (Hu2Sizes[HufN]>32) )
					  break;     // function gets decreasing
			    if(lastX>HufMax)
				  {
				  HufMax=lastX;
				  y=HufN;
				  }
			    }
			   }

			else {  /* When no optimisation, just use one fixed splitting */
			     y=1;
			     HufMax=5;
			     }

			HufN=y;		// Optimal split is stored inside 'Huffn'
			Histogram[0]=HufZero;
			for(x=Hu2Sizes[HufN]+1;x<MAX_CODES;x++) Histogram[0]+=Histogram[x];

			x=sq_huffman(Histogram,HufStruct->ln,BcountLen,Hu2Sizes[HufN]+1,ch_tab);

			/*FILE *F=fopen("c:\\temp\\huff.txt","wt");
			for(int ii=0;ii<=Hu2Sizes[HufN];ii++) fprintf(F,"%d: len:%d cod:%d\n",
			ii,HufStruct->ln[ii],HufStruct->cod[ii]);
			fclose(F); */

			  /* perform a codespace check */
			if(!HufCount(HufStruct, Hu2Sizes[HufN]+1))
			  {     //WRONG CODESPACE - BIG SHIT !!!!
			  fprintf(stderr,"Invalid Huffman codespace, plane:%d! Internal workaround.\n",i);
			  isHuffman=0;
			  goto NoHuff;
			  }

			sq_wrhufi(HufStruct,BcountLen,Hu2Sizes[HufN]+1);

			/*F=fopen("c:\\temp\\huff_.txt","wt");
			for(ii=0;ii<=Hu2Sizes[HufN];ii++) fprintf(F,"%d: len:%d cod:%d\n",ii,HufStruct->ln[ii],
			   swap_bits_order(HufStruct->cod[ii],HufStruct->ln[ii]) );
			fclose(F);*/

			isHuffman=HufMax>4;
			if(isHuffman)
				{
				if(PackSizes(NULL,HufStruct->ln,Hu2Sizes[HufN])==0) isHuffman=0;  //tohle zatim neumim
				}
			}

NoHuff:		sq_wrn(&BitStream,isHuffman,1);

		if(isHuffman)
			{
			sq_wrn(&BitStream,HufN,3);
				//Store Huffman Table
			PackSizes(&BitStream,HufStruct->ln,Hu2Sizes[HufN]);
			}
//-------end of Huffman compression------------}


		 lastX=-1;

		 for(y=0;y<PlaneX->Size2D;y++)
		    {
		    Rotm=128;
		    pb=(BYTE *)PlaneX->GetRowRaster(y)->Data1D;

		    for(x=0;x<PlaneX->GetSize1D();x++)
			{
			if((*pb & Rotm) != 0)
				{
				if(isHuffman) WriteHufNum(&BitStream,HufStruct,(long)(x)-lastX,HufN);
					 else WriteNumber(&BitStream,(long)(x)-lastX);

				lastX=x;
				}
			Rotm=Rotm >> 1;
			if(Rotm==0)
				  {
				  Rotm=128;
				  pb++;
				  }
			}
		    lastX=lastX - PlaneX->GetSize1D();
		    }
		 if(lastX<-1)
			{
			if(isHuffman) WriteHufNum(&BitStream,HufStruct,0l-lastX,HufN);
				 else WriteNumber(&BitStream,0l-lastX);
			}

		 }
		 else {
		      sq_wrn(&BitStream,0,1);
		      for(y=0;y<PlaneX->Size2D;y++)
			   {
			   x=(PlaneX->GetSize1D()-1)>>3;
			   ((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]=swap_bits_xlat[((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]];
			   sq_wrUnl(&BitStream,(BYTE *)PlaneX->GetRowRaster(y)->Data1D,PlaneX->GetSize1D());
			   ((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]=swap_bits_xlat[((BYTE *)(PlaneX->GetRowRaster(y)->Data1D))[x]];
			   }
		      }

	     }

	 TotalFlush(&BitStream);
	 free(CompBuffer);
	 }

 if(ObjLeader.size==~1)
   {
   ObjLeader.size = ftell(f) - StartPos - SizeImgObject;
   fseek(f, StartPos, SEEK_SET);
   if(SaveFtgObject(f,ObjLeader)!=SizeImgObject) return(-12); //error - cannot write object descriptor
   fseek(f, StartPos+ObjLeader.size, SEEK_SET);
   }

 if(p.Palette!=NULL)		// palete store
   {
   ObjLeader.type = 20;		//=image;
   ObjLeader.size = p.Palette->Size1D*3 + sizeof(PaletteData);
   if(SaveFtgObject(f,ObjLeader)!=SizeImgObject) return(-12); //error - cannot write object descriptor

   PaletteDesc.items = p.Palette->Size1D;
   PaletteDesc.typ = 3;
   PaletteDesc.startItem = 0;
   PaletteDesc.destPlanes = p.Raster->GetPlanes();
   if(fwrite(&PaletteDesc,sizeof(PaletteData),1,f)!=1) return -13;

   for(x=0; x<=p.Palette->Size1D; x++)
      {
      fputc(p.Palette->R(x),f);	//Red;
      fputc(p.Palette->G(x),f);	//Green;
      fputc(p.Palette->B(x),f);	//Blue;
      }
 }

 if(Histogram!=NULL) free(Histogram);
 if(HufStruct!=NULL) free(HufStruct);
 if(ch_tab!=NULL)    free(ch_tab);
 if(f!=NULL)	     fclose(f);
 if(PlaneX)	     {delete PlaneX;PlaneX=NULL;}

return(SavePictureFTG_);
}
#endif


TImageFileHandler FTG_H(".FTG",
#if SupportFTG>=4 || SupportFTG==2
	LoadPictureFTG,
#else
	NULL,
#endif
#if SupportFTG>=3
	SavePictureFTG);
#else
	NULL);
#endif


#endif
//-------------------End of FTG routines------------------




